# support code for tcl/tk socket services

namespace eval ServerSocket {
  namespace export ServerSocket SSclose SSactivate SSwrite SSbroadcast SSdrop

  variable _socket
  variable _clients
  variable _serverFunction ""
}

proc ServerSocket::ServerSocket {{port_ 20002}} {
  variable _socket
  # Open the physical socket.
  if {[catch {socket -server ServerSocket::accept $port_} _socket]} {
    return -code error
  }
  return -code ok
}

proc ServerSocket::SSdrop {sd} {
variable _clients
  unset _clients($sd)
  catch {puts $sd "close"}
  catch {close $sd}
}

proc ServerSocket::SSclose {} {
  variable _socket
  variable _clients
  variable _serverFunction

  catch {close $_socket}
  foreach sock [array names _clients] {
    ServerSocket::SSdrop $sock
  }
  set _serverFunction ""
}

proc ServerSocket::SSactivate {command_} {
variable _serverFunction
  set _serverFunction $command_
}

proc ServerSocket::SSwrite {sd_ data_} {
  if {[catch {puts $sd_ $data_} err]} {
    error $err
  }
}

proc ServerSocket::SSbroadcast {data_} {
  variable _clients
  foreach sock [array names _clients] {
    if {[catch {puts $sock $data_} err]} {
      error $err
    }
  }
}

proc ServerSocket::accept {sd_ address_ port_} {
variable _clients
  set _clients($sd_) [list $address_ $port_]
  fconfigure $sd_ -buffering line
  fileevent $sd_ readable [list ServerSocket::handleClientInput $sd_]
}

proc ServerSocket::handleClientInput {sd_} {
  variable _serverFunction
  variable _clients

  catch {gets $sd_ data} rc

  if {$rc < 0} {
    ServerSocket::SSdrop $sd_
    return
  }
  if {[string compare $data "quit"] == 0} {
    ServerSocket::SSclose
  } else {
    if {[llength $_serverFunction]} {
        eval {$_serverFunction $sd_ $data}
    } else {
      ServerSocket::SSbroadcast $data
    }
  }
}

# support code for client socket services

namespace eval ClientSocket {
  namespace export ClientSocket CSwrite CSregister CSclose
  variable _socket
  variable _clientFunction ""
}

proc ClientSocket::CSclose {} {
  variable _socket
  close $_socket
}

proc ClientSocket::CSregister {command_} {
  variable _clientFunction
  set _clientFunction $command_
}

proc ClientSocket::ClientSocket {{host_ localhost} {port_ 20002}} {
  variable _socket
  if {[catch {socket $host_ $port_} _socket]} {
    return -code error "Server DOWN"
  }
  fconfigure $_socket -buffering line
  fileevent $_socket readable ClientSocket::handleServerInput
  return $_socket
}

proc ClientSocket::CSwrite {data_} {
  variable _socket
  puts $_socket $data_
}

proc ClientSocket::handleServerInput {} {
  variable _socket
  variable _clientFunction
  if {[gets $_socket data] < 0} {
    return
  } else {
    if {[string compare $data "close"] == 0} {
      ClientSocket::CSclose
      $_clientFunction "CLOSE"
      return
    }
    if {[llength $_clientFunction]} {
      $_clientFunction $data
    } else {
      puts $data
    }
  }
}


